/* Copyright 2015-2017 Rene Widera, Alexander Grund
 *
 * This file is part of PIConGPU.
 *
 * PIConGPU is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * PIConGPU is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with PIConGPU.
 * If not, see <http://www.gnu.org/licenses/>.
 */


#pragma once

#include "simulation_defines.hpp"
#include "nvidia/rng/RNG.hpp"
#include "nvidia/rng/methods/Xor.hpp"
#include "nvidia/rng/distributions/Uniform_float.hpp"
#include "mpi/SeedPerRank.hpp"
#include "traits/GetUniqueTypeId.hpp"

namespace picongpu
{
namespace particles
{
namespace manipulators
{

namespace nvrng = nvidia::rng;
namespace rngMethods = nvidia::rng::methods;
namespace rngDistributions = nvidia::rng::distributions;

template< typename T_SpeciesType>
struct RandomPositionImpl
{
    typedef T_SpeciesType SpeciesType;
    typedef typename MakeIdentifier<SpeciesType>::type SpeciesName;

    HINLINE RandomPositionImpl(uint32_t currentStep) : isInitialized(false)
    {
        typedef typename SpeciesType::FrameType FrameType;

        GlobalSeed globalSeed;
        mpi::SeedPerRank<simDim> seedPerRank;
        seed = globalSeed() ^
               PMacc::traits::GetUniqueTypeId<FrameType, uint32_t>::uid() ^
               POSITION_SEED;
        seed = seedPerRank(seed) ^ currentStep;

        const SubGrid<simDim>& subGrid = Environment<simDim>::get().SubGrid();
        localCells = subGrid.getLocalDomain().size;
    }

    template<typename T_Particle1, typename T_Particle2>
    DINLINE void operator()(const DataSpace<simDim>& localSuperCellOffset,
                            T_Particle1& particle, T_Particle2&,
                            const bool isParticle, const bool)
    {
        typedef typename T_Particle1::FrameType FrameType;

        if (!isInitialized)
        {
            /** @todo: it is a wrong assumption that the threadIdx can be used to
             * define the cell within the superCell. This is only allowed if we not
             * use alpaka. We need to distinguish between manipulators those are working on the
             * cell domain and on the particle domain.
             */
            const DataSpace<simDim > threadIndex(threadIdx);
            const uint32_t cellIdx = DataSpaceOperations<simDim>::map(
                                                                      localCells,
                                                                      localSuperCellOffset + threadIndex);
            rng = nvrng::create(rngMethods::Xor(seed, cellIdx), rngDistributions::Uniform_float());
            isInitialized = true;
        }
        if (isParticle)
        {
            floatD_X tmpPos;

            for (uint32_t d = 0; d < simDim; ++d)
                tmpPos[d] = rng();

            particle[position_] = tmpPos;
        }
    }

private:
    typedef PMacc::nvidia::rng::RNG<rngMethods::Xor, rngDistributions::Uniform_float> RngType;
    RngType rng;
    bool isInitialized;
    uint32_t seed;
    DataSpace<simDim> localCells;
};

} //namespace manipulators
} //namespace particles
} //namespace picongpu
